'''
Create a new document with highlights of the current document.
Can be used by adding the following to your prefs_user.config:

new_command _extract_highlights python -m sioyek.extract_highlights "%{sioyek_path}" "%{local_database}" "%{shared_database}" "%{file_path}" %{zoom_level}

'''
import os
import sys
import pathlib
import sqlite3
from copy import copy
import hashlib
import subprocess

from .sioyek import Sioyek, clean_path
from PyPDF2 import PdfWriter, PdfReader

LOCAL_DATABASE_FILE = None
SHARED_DATABASE_FILE = None

sioyek = None
SELECT_DOCUMENT_HIGHLIGHTS_QUERY = "SELECT * from highlights where document_path='{}'"
SELECT_HASH_QUERY = "SELECT * from document_hash where path='{}'"
INSERT_DOCUMENT_HASH = "INSERT INTO document_hash (path, hash) VALUES ('{}', '{}')"
PORTAL_INSERT_QUERY = "INSERT INTO links (src_document, dst_document, src_offset_y, dst_offset_x, dst_offset_y, dst_zoom_level) VALUES ('{}', '{}', {}, {}, {}, {})"
PORTAL_DELETE_QUERY = "DELETE FROM links WHERE src_document='{}'"
UPDATE_DOCUMENT_HASH_QUERY = "UPDATE document_hash SET hash='{}' WHERE path='{}'"
INSERT_NEW_DOCUMENT_HASH_QUERY = "INSERT INTO document_hash (path, hash) VALUES ('{}', '{}')"

SIOYEK_PATH = None

def md5_hash(file_path):
    file_hash = hashlib.md5()

    with open(file_path, "rb") as f:
        chunk = f.read(8192)
        while chunk:
            file_hash.update(chunk)
            chunk = f.read(8192)
            
    return file_hash.hexdigest()

def delete_prev_portals_of_document(shared_cursor, document_hash):
    query = PORTAL_DELETE_QUERY.format(document_hash)
    shared_cursor.execute(query)

def get_hash_with_path(local_cursor, file_path):
    results = local_cursor.execute(SELECT_HASH_QUERY.format(file_path)).fetchall()
    if len(results) > 0:
        return results[0][-1]
    else:
        return None

def update_document_hash(local_cursor, document_path, document_hash):
    prev_hash = get_hash_with_path(local_cursor, document_path)
    if prev_hash != None and prev_hash != document_hash:
        query = UPDATE_DOCUMENT_HASH_QUERY.format(document_hash, document_path)
        local_cursor.execute(query)

    if prev_hash == None:
        query = INSERT_NEW_DOCUMENT_HASH_QUERY.format(document_path, document_hash)
        local_cursor.execute(query)


if __name__ == '__main__':

    doc_path = None
    new_file_path = None

    if len(sys.argv) > 1:
        SIOYEK_PATH = clean_path(sys.argv[1])
        LOCAL_DATABASE_FILE = clean_path(sys.argv[2])
        SHARED_DATABASE_FILE = clean_path(sys.argv[3])
        doc_path = clean_path(sys.argv[4])
        zoom_level = float(sys.argv[5])
        doc_dir = os.path.dirname(doc_path)
        doc_base_file_name = os.path.basename(doc_path).split('.')[0]
        new_file_name = doc_base_file_name + '_highlights.pdf'
        new_file_path = str(pathlib.Path(doc_dir) / new_file_name).replace('\\', '/')
    sioyek = Sioyek(SIOYEK_PATH, LOCAL_DATABASE_FILE, SHARED_DATABASE_FILE)

    # open sqlite3 database
    local_database = sqlite3.connect(LOCAL_DATABASE_FILE)
    shared_database = sqlite3.connect(SHARED_DATABASE_FILE)

    doc = sioyek.get_document(doc_path)
    document_hash = doc.get_hash()
    document_highlights = doc.get_highlights()

    highlight_bounding_boxes = [doc.get_highlight_bounding_box(highlight.get_begin_abs_pos(), highlight.get_end_abs_pos()) for highlight in document_highlights]

    highlight_bounding_boxes = sorted(highlight_bounding_boxes, key=lambda x: x[1])
    
    pdf_reader = PdfReader(doc_path)
    pdf_writer = PdfWriter()

    new_document_offsets = []

    offset = 0
    blank_height = 50
    for bounding_box, page_number in highlight_bounding_boxes:
        pdf_page = copy(pdf_reader.pages[page_number])
        h = int(pdf_page.mediabox.height)


        margin = 20
        old_mediabox = pdf_page.mediabox.copy()
        pdf_page.mediabox.upper_left = (bounding_box[0], h - bounding_box[1] + margin)
        pdf_page.mediabox.lower_right = (bounding_box[2],  h - bounding_box[3] - margin)
        pdf_writer.add_page(pdf_page)
        pdf_writer.add_blank_page(height=blank_height, width=pdf_page.mediabox.width)

        new_document_offsets.append(offset)
        offset += int(pdf_page.mediabox.height) + blank_height

    with open(new_file_path, 'wb') as f:
        pdf_writer.write(f)

    new_file_hash = md5_hash(new_file_path)

    local_cursor = local_database.cursor()
    shared_cursor = shared_database.cursor()
    local_cursor.execute('begin')
    shared_cursor.execute('begin')
    try:

        update_document_hash(local_cursor, new_file_path, new_file_hash)
        delete_prev_portals_of_document(shared_cursor, new_file_hash)
        for src_offset, highlight in zip(new_document_offsets, document_highlights):
            dst_y_offset = (highlight.selection_begin[1] + highlight.selection_end[1]) / 2
            dst_zoom_level = zoom_level / 2
            dst_document = document_hash
            src_document = new_file_hash
            query = PORTAL_INSERT_QUERY.format(src_document, dst_document, src_offset, 0, dst_y_offset, dst_zoom_level)
            shared_cursor.execute(query)

        local_cursor.execute('commit')
        shared_cursor.execute('commit')
    except sqlite3.Error as e:
        print(str(e))
        local_cursor.execute('rollback')
        shared_cursor.execute('rollback')
    
    subprocess.run([SIOYEK_PATH, new_file_path, '--new-window'])
    subprocess.run([SIOYEK_PATH, '--execute-command', 'reload'])

    local_database.close()
    shared_database.close()



'''
Allows sioyek to automatically download papers from scihub by clicking on their names.

Here is an example `prefs_user.config` file which uses this script:

    new_command _download_paper_under_cursor python -m sioyek.paper_downloader download "%{sioyek_path}" "%{paper_name}"
    control_click_command _download_paper_under_cursor

Now, you can control+click on paper names to download them and open them in sioyek.

This script can also be used to copy the bibtex of paper under cursor:
    new_command _copy_bibtex python -m sioyek.paper_downloader copy download "%{sioyek_path}" "%{paper_name}"

'''

# where to put downloaded papers, if it is None, we use a default data path
PAPERS_FOLDER_PATH = None
SIOYEK_PATH = None
PYTHON_EXECUTABLE = 'python'

sioyek = None

import os
import pathlib
import subprocess
import sys
import json
import time
import regex
import urllib.request
import fitz

import pyperclip
from appdirs import user_data_dir
from habanero import Crossref
from PyPaperBot.__main__ import start as start_paper_download

from .sioyek import Sioyek, clean_path


def clean_paper_name(paper_name):
    first_quote_index = -1
    last_quote_index = -1

    try:
        first_quote_index = paper_name.index('"')
        last_quote_index = paper_name.rindex('"')
    except ValueError as e:
        pass

    if first_quote_index != -1 and last_quote_index != -1 and (last_quote_index - first_quote_index) > 10:
        return paper_name[first_quote_index + 1:last_quote_index]

    paper_name = paper_name.strip()
    if paper_name.endswith('.'):
        paper_name = paper_name[:-1]

    return paper_name

def is_file_a_paper_with_name(file_name, paper_name):
    # check if the downloaded file's title matches with the query
    doc = fitz.open(file_name)
    page_text = doc.get_page_text(0)
    doc.close()
    if regex.search('(' + regex.escape(paper_name) + '){e<=6}', page_text, flags=regex.IGNORECASE):
        return True
    return False


class ListingDiff():
    def __init__(self, path):
        self.path = path

    def get_listing(self):
        return set(os.listdir(self.path))

    def __enter__(self):
        self.listing = self.get_listing()
        return self
    
    def new_files(self):
        return list(self.get_listing().difference(self.listing))
    
    def new_pdf_files(self):
        return [f for f in self.new_files() if f.endswith('.pdf')]
    
    def reset(self):
        self.listing = self.get_listing()

    def __exit__(self, exc_type, exc_value, exc_traceback):
        pass

def get_cached_doi_map():
    json_path = get_papers_folder_path() / 'doi_map.json'
    if os.path.exists(json_path):
        with open(json_path, 'r') as infile:
            return json.load(infile)
    else:
        return dict()

def write_cache_doi_map(doi_map):
    json_path = get_papers_folder_path() / 'doi_map.json'
    with open(json_path, 'w') as outfile:
        json.dump(doi_map, outfile)


def get_papers_folder_path_():
    if PAPERS_FOLDER_PATH != None:
        return pathlib.Path(PAPERS_FOLDER_PATH)
    APPNAME = 'sioyek_papers'
    return pathlib.Path(user_data_dir(APPNAME, False))

def get_papers_folder_path():
    path = get_papers_folder_path_()
    path.mkdir(parents=True, exist_ok=True)
    return path

def get_doi_with_name(paper_name):
    crossref = Crossref()
    response = crossref.works(query=paper_name)
    if len(response['message']['items']) == 0:
        return None
    return response['message']['items'][0]['DOI']

def download_paper_with_doi(doi_string, paper_name, doi_map):
    download_dir = get_papers_folder_path()
    method_args_and_kwargs = [
        ("google scholar", (paper_name, 1, [1], download_dir, None), {}),
        ("scihub", ("", None, None, download_dir, None), {'DOIs': [doi_string]}),
    ]

    with ListingDiff(download_dir) as listing_diff:

        ignored_files = []

        for method_name, method_args, method_kwargs in method_args_and_kwargs:
            sioyek.set_status_string('trying to download "{}" from {}'.format(paper_name, method_name))
            try:
                start_paper_download(*method_args, **method_kwargs)
            except Exception as e:
                sioyek.set_status_string('error in download from {}'.format(method_name))
            pdf_files = listing_diff.new_pdf_files()
            if len(pdf_files) > 0:
                returned_file = download_dir / pdf_files[0]
                if is_file_a_paper_with_name(str(returned_file), paper_name):
                    doi_map[doi_string] = str(returned_file)
                    write_cache_doi_map(doi_map)
                    return returned_file
                else:
                    ignored_files.append(returned_file)
                    listing_diff.reset()

        if len(ignored_files) > 0:
            sioyek.set_status_string('could not find a suitable paper, this is a throw in the dark')
            return ignored_files[0]

    return None

def get_paper_file_name_with_doi_and_name(doi, paper_name):

    doi_map = get_cached_doi_map()
    if doi in doi_map.keys():
        if os.path.exists(doi_map[doi]):
            return doi_map[doi]
    return download_paper_with_doi(doi, paper_name, doi_map)


def get_bibtex(doi):
    BASE_URL = 'http://dx.doi.org/'
    url = BASE_URL + doi
    req = urllib.request.Request(url)
    req.add_header('Accept', 'application/x-bibtex')
    with urllib.request.urlopen(req) as f:
        bibtex = f.read().decode()
        return bibtex

if __name__ == '__main__':

    mode = sys.argv[1]
    SIOYEK_PATH = clean_path(sys.argv[2])
    sioyek = Sioyek(SIOYEK_PATH)
    paper_name = clean_paper_name(sys.argv[3])

    sioyek.set_status_string('finding doi ...')
    try:

        doi = get_doi_with_name(sys.argv[3])
        if doi:
            if mode == 'download':
                # show_status('downloading doi: {}'.format(doi))
                file_name = get_paper_file_name_with_doi_and_name(doi, paper_name)
                sioyek.clear_status_string()
                if file_name:
                    subprocess.run([SIOYEK_PATH, str(file_name), '--new-window'])
            else:
                bibtex = get_bibtex(doi)
                pyperclip.copy(bibtex)
                sioyek.clear_status_string()

        else:
            sioyek.set_status_string('doi not found')
            time.sleep(5)
            sioyek.clear_status_string()
    except Exception as e:
        sioyek.set_status_string('error: {}'.format(str(e)))
        time.sleep(5)
        sioyek.clear_status_string()


import sys
from googletrans import Translator

from .sioyek import Sioyek, clean_path

if __name__ == '__main__':
    sioyek_path = clean_path(sys.argv[1])
    text = sys.argv[2]
    sioyek = Sioyek(sioyek_path)
    translator = Translator()
    translation = translator.translate(text, dest='en')
    sioyek.set_status_string(translation.text)